import re
import sys

from django.apps import apps
from django.utils.translation import gettext_lazy as _


__all__ = (
    'MetaViewWork',
    'resolver',
)


class MetaViewWork(type):
    @staticmethod
    def _get_class_app(cls):
        app_name = ''
        for module in cls.__module__.split('.'):
            app_name += module
            if f'{app_name}.apps' in sys.modules and apps.is_installed(app_name):
                break
            app_name += '.'
        try:
            return app_name and next(ac for ac in apps.app_configs.values() if ac.name == app_name) or None
        except (StopIteration, IndexError):
            ...
        return None

    def __new__(cls, name, bases, attrs):
        new_cls = super(MetaViewWork, cls).__new__(cls, name, bases, attrs)

        if new_cls.__dict__.get('vw_abstract', False):
            return new_cls

        vw_name = new_cls.__dict__.get('vw_name', getattr(new_cls, 'vw_name', ''))
        vw_verbose_name = getattr(new_cls, 'vw_verbose_name', '')
        cls_name = filter(bool, re.split('(?=[A-Z])', new_cls.__name__))
        if not (vw_name or vw_verbose_name):
            return new_cls
        elif not vw_name:
            vw_name = '_'.join(cls_name).lower()
            new_cls.vw_name = vw_name
        elif not vw_verbose_name:
            new_cls.vw_verbose_name = _(' '.join(cls_name))

        app = cls._get_class_app(new_cls)
        if app:
            vw_prefix = getattr(new_cls, 'vw_prefix', '').strip()
            new_cls.vw[app.label][vw_name] = new_cls
            new_cls.vw[app.label][f'{vw_prefix}{vw_name}'] = new_cls

        return new_cls


def resolver(*classes, **options):
    """Metaclass resolver

    Example:
        SomeClass(resolver(SomeMixin, SomeBaseClass, ..., some_generic_field=..., ...)):
            ...

    :param classes:
    :param options:
    :return: generic metaclass
    """
    is_not_type = lambda cls: cls is not type
    generic_name = lambda gc: '_'.join(c.__name__ for c in gc)
    metaclass = tuple(set(filter(is_not_type, map(type, classes))))
    metaclass = (
        metaclass[0] if len(metaclass) == 1
        else type(generic_name(metaclass), metaclass, options)
    )
    return metaclass(generic_name(classes), classes, options)



import sys

from django.apps import apps as real_apps
from django.db import migrations, models

from viewwork import BaseViewWork


def fix_view_name(apps, schema_editor):
    Menu = apps.get_model('viewwork', 'Menu')

    for app_label, values in BaseViewWork.vw.items():
        app = real_apps.get_app_config(app_label)
        urls = sys.modules[f'{app.module.__name__}.urls']
        namespace = getattr(urls, 'app_name', None) or app.module.__name__
        for item in Menu.objects.filter(view__in=values.keys()):
            item.view = f'{namespace}:{item.view}'
            item.save(update_fields=('view',))


def remove_app_label(apps, schema_editor):
    Menu = apps.get_model('viewwork', 'Menu')

    for item in Menu.objects.filter(view__icontains=':'):
        item.view = item.view.split(':')[1]
        item.save(update_fields=('view',))


class Migration(migrations.Migration):
    dependencies = [
        ('viewwork', '0004_enabled_hidden'),
    ]

    operations = [
        migrations.AlterField(
            model_name='menu',
            name='view',
            field=models.CharField(db_index=True, default='', max_length=125, verbose_name='View'),
        ),

        migrations.RunPython(fix_view_name, reverse_code=remove_app_label),
    ]



from collections import defaultdict

from django.views.generic import TemplateView

from .metaclass import resolver as metaclass_resolver, MetaViewWork


__all__ = (
    'default_app_config',
    'BaseViewWork',
)


default_app_config = 'viewwork.apps.ViewWorkConfig'


class BaseViewWork(TemplateView, metaclass=MetaViewWork):
    vw: dict = defaultdict(dict)
    vw_abstract: bool = True
    vw_prefix: str = ''



# Generated by Django 3.1.3 on 2020-11-22 19:41

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Menu',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80, verbose_name='Menu item name')),
                ('view', models.CharField(default='', max_length=80, verbose_name='View')),
                ('sort_order', models.SmallIntegerField(default=0, verbose_name='Sort order')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='menu_parent', to='viewwork.menu', verbose_name='Parent menu item')),
            ],
            options={
                'verbose_name': 'Menu item',
                'verbose_name_plural': 'Menu items',
            },
        ),
    ]



from importlib import import_module
from pathlib import Path
import pkgutil

from django.apps import AppConfig, apps
from django.urls import path, include
from django.utils.translation import gettext_lazy as _

from . import BaseViewWork, settings


__all__ = (
    'ViewWorkConfig',
)


class ViewWorkConfig(AppConfig):
    name: str = 'viewwork'
    verbose_name: str = _('ViewWork')

    def collect_views(self, app: AppConfig):
        app_name = app.module.__name__
        import_module(f'{app_name}.views')
        for entry in pkgutil.walk_packages([str(Path(app.path) / 'views')], f'{app_name}.views.'):
            import_module(entry.name)

    def collect_urls(self, app: AppConfig):
        urls = import_module(f'{app.module.__name__}.urls')
        app_urls_namespace = getattr(urls, 'app_name', None)
        urlpatterns = [
            path(f'{name}/', view_class.as_view(), name=name)
            for name, view_class in BaseViewWork.vw[app.label].items()
        ]
        if app_urls_namespace or not settings.USE_APP_NAMESPACE:
            urls.urlpatterns += urlpatterns
        else:
            urls.urlpatterns += [path('', include((urlpatterns, app.module.__name__)))]

    def ready(self):
        for app in apps.get_app_configs():
            if not getattr(app, 'vw_collect', False):
                continue

            self.collect_views(app)
            self.collect_urls(app)



from django.conf import settings
from django.core.signals import setting_changed


ADD_USER_MENU = getattr(settings, 'VW_ADD_USER_MENU', True)
ADD_ADMIN_SITE = getattr(settings, 'VW_ADD_ADMIN_SITE', True)
ADD_USER_LOGOUT = getattr(settings, 'VW_ADD_USER_LOGOUT', True)
ADMIN_SITE_URL = getattr(settings, 'VW_ADMIN_SITE_URL', '/admin/')
VW_PREFIX = getattr(settings, 'VW_PREFIX', '')
REMOVE_EMPTY_ITEM = getattr(settings, 'VW_REMOVE_EMPTY_ITEM', True)
USE_APP_NAMESPACE = getattr(settings, 'VW_USE_APP_NAMESPACE', True)


def reload_settings(setting, value, **kwargs):
    if setting == 'VW_REMOVE_EMPTY_ITEM':
        global REMOVE_EMPTY_ITEM
        REMOVE_EMPTY_ITEM = value
    elif setting == 'VW_PREFIX':
        global VW_PREFIX
        VW_PREFIX = value


setting_changed.connect(reload_settings)



from operator import attrgetter, itemgetter
import sys

from django import forms
from django.apps import apps
from django.conf import settings
from django.db.models import BLANK_CHOICE_DASH
from django.utils.translation import gettext_lazy as _, get_language

from modeltrans.utils import build_localized_fieldname
try:
    from easy_select2 import Select2
except (ImportError, ModuleNotFoundError):
    Select2 = forms.Select

from . import BaseViewWork, settings
from .models import Menu


__all__ = (
    'MenuAdminForm',
)


class MenuAdminForm(forms.ModelForm):
    view = forms.ChoiceField(label=_('View'), required=False, widget=Select2())

    def __init__(self, *args, **kwargs):
        super(MenuAdminForm, self).__init__(*args, **kwargs)

        view_choices= []
        vw_prefix = attrgetter('vw_prefix')
        item1 = itemgetter(1)
        for app_label, opts in BaseViewWork.vw.items():
            app = apps.get_app_config(app_label)
            urls = sys.modules[f'{app.module.__name__}.urls']
            namespace = getattr(urls, 'app_name', None) or app.module.__name__
            namespace = f'{namespace}:' if settings.USE_APP_NAMESPACE else ''
            views = filter(lambda v: not vw_prefix(v[1]) or not v[0].startswith(vw_prefix(v[1])), opts.items())
            views = list(map(item1, views))
            options = [(f'{namespace}{v.vw_name}', f'{v.vw_verbose_name}: {v.vw_name}') for v in views]
            options.sort(key=item1)
            view_choices.append((app.verbose_name, options))

        view_choices.sort(key=lambda opt: opt[0])
        view_choices.insert(0, BLANK_CHOICE_DASH[0])

        self.fields['view'].choices = view_choices
        self.fields['parent'].queryset = Menu.objects.filter(view='').order_by('name_i18n')
        self.fields['name_i18n'].required = True

    class Meta:
        model = Menu
        fields = '__all__'
        exclude = ('name',)
        widgets = {
            'parent': Select2()
        }

    @staticmethod
    def _strip(value):
        return (value or '').strip()

    def clean_name_i18n(self):
        return self._strip(self.cleaned_data.get('name_i18n'))

    def clean(self):
        cleaned_data = super().clean()
        for lang in {get_language(), settings.LANGUAGE_CODE}:
            field_name = build_localized_fieldname('name', lang)
            cleaned_data[field_name] = self._strip(cleaned_data.get(field_name)) or cleaned_data.get('name_i18n', '')
        return cleaned_data



# Generated by Django 3.1.13 on 2021-10-31 16:59

from django.db import migrations, models


def set_defaults(apps, schema_editor):
    Menu = apps.get_model('viewwork', 'Menu')

    Menu.objects.all().update(hidden=False)


class Migration(migrations.Migration):

    dependencies = [
        ('viewwork', '0003_menu_i18n'),
    ]

    operations = [
        migrations.AddField(
            model_name='menu',
            name='enabled',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Enabled item'),
        ),
        migrations.AddField(
            model_name='menu',
            name='hidden',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Hidden item'),
        ),

        migrations.RunPython(set_defaults, reverse_code=migrations.RunPython.noop)
    ]



import sys

from django.apps import apps
from django.core.management import BaseCommand

from viewwork import BaseViewWork
from viewwork.models import Menu


class Command(BaseCommand):
    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument('action', action='store', type=str, choices=['add', 'delete'])

    def add(self):
        for app_label, values in BaseViewWork.vw.items():
            app = apps.get_app_config(app_label)
            urls = sys.modules[f'{app.module.__name__}.urls']
            namespace = getattr(urls, 'app_name', None) or app.module.__name__
            for item in Menu.objects.filter(view__in=values.keys()):
                item.view = f'{namespace}:{item.view}'
                item.save(update_fields=('view',))

    def delete(self):
        for item in Menu.objects.filter(view__icontains=':'):
            item.view = item.view.split(':')[1]
            item.save(update_fields=('view',))

    def handle(self, *args, **options):
        if options['action'] == 'add':
            self.add()
        elif options['action'] == 'delete':
            self.delete()



from django.views.generic import View
from django.http import JsonResponse
from django.urls import reverse, resolve, NoReverseMatch, Resolver404
from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _
from django.db.models import Q

try:
    from guardian.shortcuts import get_objects_for_user
except (ImportError, ModuleNotFoundError):
    get_objects_for_user = None

from .models import Menu
from .utils import build_nested_tree
from . import settings


__all__ = (
    'MenuView',
    'menu',
)


User = get_user_model()


class MenuView(View):
    http_method_names = ['get']

    def get_queryset(self):
        if not get_objects_for_user or self.request.user.is_superuser:
            menu = Menu.objects.all()
        else:
            menu_items = get_objects_for_user(
                self.request.user, 'viewwork.view_menu'
            ).values_list('pk', flat=True)
            menu = Menu.objects.filter(Q(pk__in=menu_items) | Q(view=''))
        return menu.filter(enabled=True, hidden=False).values(
            'id', 'parent_id', 'name_i18n', 'view'
        ).order_by('parent_id', 'sort_order', 'name_i18n')

    def patch_tree(self, tree):
        if settings.ADD_USER_MENU and self.request.user.is_authenticated:
            tree.append({
                'id': '-2',
                'parent_id': None,
                'name': getattr(self.request.user, User.USERNAME_FIELD),
                'url': '',
                'items': [],
            })
            if settings.ADD_ADMIN_SITE and self.request.user.is_staff:
                tree[-1]['items'].append({
                    'id': '-3', 'parent_id': '-2', 'name': _('Admin site'),
                    'url': settings.ADMIN_SITE_URL, '_url': settings.ADMIN_SITE_URL,
                })
            if settings.ADD_USER_LOGOUT:
                tree[-1]['items'].append({
                    'id': '-4', 'parent_id': '-2', 'name': _('Logout'),
                    'url': reverse('vw_logout'), '_url': reverse('vw_logout'), 'target': '_self',
                })
        return tree

    def get(self, request, *args, **kwargs):
        menu_items = self.get_queryset()
        data = []
        for item in menu_items:
            if item['view']:
                try:
                    item['url'] = reverse(item['view'])
                except NoReverseMatch:
                    try:
                        resolve(item['view'])
                        item['url'] = item['view']
                    except Resolver404:
                        continue
                try:
                    item['_url'] = reverse(f"{settings.VW_PREFIX}{item['view']}")
                except NoReverseMatch:
                    item['_url'] = item['url']
            if 'name_i18n' in item:
                item['name'] = item.pop('name_i18n', '')
            data.append(item)
        tree = build_nested_tree(data)
        tree = self.patch_tree(tree)
        return JsonResponse(tree, safe=False)


menu = MenuView.as_view()



from django.urls import path
from django.contrib.auth.views import logout_then_login

from viewwork.views import menu


urlpatterns = [
    path('', menu),
    path('logout/', logout_then_login, name='vw_logout'),
]



from . import settings


def empty_items(raw_items, nodes, parent_name='parent_id', id_name='id'):
    empty_parents = {item[id_name] for item in raw_items if not item['view']}
    parents = {item[parent_name] for item in raw_items if item['view'] and item[parent_name]}
    for p in parents:
        while p is not None:
            empty_parents -= {p}
            p = nodes[p].get(parent_name)
    return empty_parents


def build_nested_tree(
        raw_items,
        parent_name='parent_id',
        children_name='items',
        id_name='id',
        parent_null=None
):
    REMOVE_EMPTY_ITEM = getattr(settings, 'REMOVE_EMPTY_ITEM', True)
    nodes = {item[id_name]: item for item in raw_items}
    remove_parents = empty_items(raw_items, nodes, parent_name=parent_name, id_name=id_name) if REMOVE_EMPTY_ITEM else {}
    nested_tree = []
    parent_null = parent_null or 0
    for item in raw_items:
        if item[id_name] in remove_parents:
            continue
        parent_id = item[parent_name] or parent_null
        node = nodes[item[id_name]]
        if parent_id == parent_null:
            nested_tree.append(node)
        else:
            parent = nodes[parent_id]
            if not children_name in parent:
                parent[children_name] = []
            parent[children_name].append(node)
    return nested_tree



# Generated by Django 3.1.13 on 2021-10-16 20:42

from django.db import migrations
import modeltrans.fields


class Migration(migrations.Migration):

    dependencies = [
        ('viewwork', '0002_auto_20210709_1622'),
    ]

    operations = [
        migrations.AddField(
            model_name='menu',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=('name',), required_languages=(), virtual_fields=True),
        ),
    ]



from modeltrans.fields import TranslationField

from django.db import models
from django.utils.translation import gettext_lazy as _


__all__ = (
    'Menu',
)


class Menu(models.Model):
    parent = models.ForeignKey(
        'self', related_name='%(class)s_parent', verbose_name=_('Parent menu item'),
        null=True, blank=True, on_delete=models.PROTECT)
    name = models.CharField(_('Menu item name'), max_length=80)
    view = models.CharField(_('View'), max_length=125, default='', db_index=True)
    sort_order = models.SmallIntegerField(_('Sort order'), default=0)
    hidden = models.BooleanField(_('Hidden item'), default=True, db_index=True)
    enabled = models.BooleanField(_('Enabled item'), default=True, db_index=True)

    i18n = TranslationField(fields=('name',))

    class Meta:
        verbose_name = _('Menu item')
        verbose_name_plural = _('Menu items')
        permissions = [
            ('view_full_access', _('View full access')),
            ('view_read_only', _('View read only')),
        ]

    def __str__(self):
        return self.name_i18n



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('viewwork', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='menu',
            options={'permissions': [('view_full_access', 'View full access'), ('view_read_only', 'View read only')], 'verbose_name': 'Menu item', 'verbose_name_plural': 'Menu items'},
        ),
    ]



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.md').read()
requirements = open('requirements/base.txt').readlines()
require_select2 = open('requirements/select2.txt').readlines()
require_all = requirements + require_select2

setup(
    name='django-viewwork',
    version='0.6.2',
    description='Collect views, generate urls and create menu',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Sergei Pikhovkin',
    author_email='s@pikhovkin.ru',
    url='https://github.com/pikhovkin/django-viewwork',
    packages=[
        'viewwork',
    ],
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        'all': require_all,
        'select2': require_select2,
    },
    python_requires='>=3.8.*, <4.0.*',
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    keywords=[
        'automation',
        'django',
        'views',
        'menu',
        'urls',
        'collectionview',
        'django-views',
        'django-urls',
        'autogeneration',
        'django-menu',
    ]
)



from django.contrib import admin
from django.utils.translation import gettext_lazy as _, get_language

from modeltrans.utils import build_localized_fieldname
try:
    from guardian.admin import GuardedModelAdmin as ModelAdmin
except (ImportError, ModuleNotFoundError):
    ModelAdmin = admin.ModelAdmin

from .models import Menu
from .forms import MenuAdminForm


class ParentListFilter(admin.SimpleListFilter):
    title = _('Parent menu item')
    parameter_name = 'parent'

    def lookups(self, request, model_admin):
        return [
            (m.id, m.name)
            for m in Menu.objects.filter(
                view=''
            ).select_related('parent').order_by('name_i18n')
        ]

    def queryset(self, request, queryset):
        val = self.value()
        if val:
            return queryset.filter(parent=val)


class MenuAdmin(ModelAdmin):
    form = MenuAdminForm
    list_display = ('parent', 'name_i18n', 'view', 'sort_order',)
    list_editable = ('sort_order',)
    list_display_links = ('name_i18n',)
    list_filter = (ParentListFilter,)
    search_fields = ('name_i18n', 'view',)
    ordering = ('parent', 'name_i18n', 'sort_order',)
    exclude = ('name',)

    def get_exclude(self, request, obj=None):
        excludes = set(super().get_exclude(request) or [])
        excludes.add(build_localized_fieldname('name', get_language()))
        return list(excludes)


admin.site.register(Menu, MenuAdmin)
